from __future__ import annotations

from dataclasses import dataclass
from inspect import get_annotations
from types import NoneType
from typing import Any, Callable, Generic, cast, overload

from .ctypes import CFUNCTYPE, FuncPointerType, StrType, c_double, c_int, c_void_p
from .string import String
from .types import CallingConvention, CDataBase, F, FuncPointer, P, Pointer, R, T

__all__ = [
    '_protected_keys',

    'ord_if_char',

    'normalize_cfunc', 'normalize_ctype',

    'as_cfunc', 'wrap_func_pointer',

    'with_signature',
]


_protected_keys = {
    '_fields_',
}


@overload
def ord_if_char(value: str | bytes) -> int:  # type: ignore
    ...


@overload
def ord_if_char(value: T) -> T:
    ...


def ord_if_char(value: T) -> T | int:
    if isinstance(value, (bytes, str)):
        return ord(value)

    return value


@dataclass
class NormalizedFunction(Generic[P, R]):
    func: Callable[P, R]
    name: str
    oname: str | None
    args_types: list[type[CDataBase]]
    oargs_types: list[type[CDataBase]] | None
    res_type: type[CDataBase]
    ores_type: type[CDataBase] | None
    cconv: CallingConvention


def normalize_cfunc(
    func: Callable[P, R], name: str | None = None, def_cconv: CallingConvention = CallingConvention.C
) -> NormalizedFunction[P, R]:
    if '__wrapped__' in func.__dir__():
        old_dict = func.__dict__.copy()
        func = func.__wrapped__  # type: ignore
        func.__dict__ |= old_dict | func.__dict__

    if name is None:
        name = func.__name__

    arguments = get_annotations(func, eval_str=True)
    return_type = arguments.pop('return')
    args_types_raw = list(arguments.values())

    oname = func.__dict__.get('__ctdffi_oname__', None)
    cconv = func.__dict__.get('__ctdffi_cconv__', def_cconv)

    res_type, *args_types = [Pointer.normalize(val) for val in (return_type, *args_types_raw)]

    ores_type = func.__dict__.get('__ctdffi_ores_type__', None)
    oargs_types = func.__dict__.get('__ctdffi_oargs_types__', None)

    if ores_type is not None:
        ores_type = Pointer.normalize(ores_type)

    if oargs_types is not None:
        oargs_types = [Pointer.normalize(val) for val in oargs_types]

    return NormalizedFunction(func, name, oname, args_types, oargs_types, res_type, ores_type, cconv)


@overload
def as_cfunc(func: Callable[P, R], name: str | None = None) -> type[FuncPointer[P, R]]:
    ...


@overload
def as_cfunc(func: NormalizedFunction[P, R], name: str | None = None) -> type[FuncPointer[P, R]]:
    ...


def as_cfunc(func: Callable[P, R] | NormalizedFunction[P, R], name: str | None = None) -> type[FuncPointer[P, R]]:
    if not isinstance(func, NormalizedFunction):
        func = normalize_cfunc(func, name)

    return CFUNCTYPE(  # type: ignore
        func.ores_type or func.res_type, *(func.oargs_types or func.args_types)
    )


def with_signature(
    args_types: list[type[CDataBase]] | None = None,
    res_type: type[CDataBase] | None = None,
    name: str | None = None, **kwargs: type[CDataBase]
) -> Callable[[F], F]:
    def wrapper(func: F) -> F:
        nonlocal args_types, res_type, name, kwargs

        if name is not None:
            func.__dict__.__setitem__('__ctdffi_oname__', name)

        annotations = get_annotations(func, eval_str=True)

        if kwargs:
            if (empty_args := not args_types):
                args_types = []
            for i, (param_name, otype) in enumerate(annotations.items()):
                if param_name == 'return':
                    continue

                arg_type = kwargs.get(param_name, otype)

                if empty_args:
                    args_types.append(arg_type)
                else:
                    try:
                        args_types[i] = arg_type
                    except KeyError:
                        raise ValueError(
                            'override: Not enough arguments specified for args_types!'
                        )
        elif args_types:
            if len(args_types) != len(annotations) - ('return' in annotations):
                raise ValueError(
                    'override: Arguments number mismatch for args_types!'
                )

        if args_types is not None:
            func.__dict__.__setitem__('__ctdffi_oargs_types__', args_types)

        if res_type is not None:
            func.__dict__.__setitem__('__ctdffi_ores_type__', res_type)

        return func

    return wrapper


def wrap_func_pointer(
    func_ptr: FuncPointerType, name: str | None = None,
    def_cconv: CallingConvention = CallingConvention.C
) -> Callable[[Callable[P, R]], FuncPointer[P, R]]:
    def wrapper(func: Callable[P, R]) -> FuncPointer[P, R]:
        norm = normalize_cfunc(func, name, def_cconv)

        func_pointer = func_ptr
        func_pointer.argtypes = norm.args_types
        func_pointer.restype = norm.res_type

        return func_pointer  # type: ignore

    return wrapper


_normalization_map = {
    NoneType: c_void_p,
    float: c_double,
    int: c_int,
    bool: c_int,
    str: String,
    StrType: String
}


def normalize_ctype(value: Any) -> type[CDataBase]:
    return cast(type[CDataBase], _normalization_map.get(value, value))



from __future__ import annotations

from abc import abstractmethod
from ctypes import POINTER, Structure, c_void_p
from enum import Enum
from typing import TYPE_CHECKING, Any, Callable, Generic, ParamSpec, Sequence, TypeVar

if TYPE_CHECKING:
    from ctypes import _CData as CDataBase
    from ctypes import _Pointer  # type: ignore
    from ctypes import _StructUnionMeta as StructMetaBase
else:
    _Pointer = Generic
    PointerBase = object
    StructMetaBase = type(Structure)
    CDataBase = Structure.__bases__[0]


__all__ = [
    'MetaClassDictBase',
    'StructMetaBase',
    'CDataBase',
    'Pointer', 'FuncPointer',
    'T', 'F', 'P', 'R', 'C_T', 'Self',
    'CallingConvention'
]


T = TypeVar('T')
F = TypeVar('F', bound=Callable[..., Any])
C_T = TypeVar('C_T')
C_TB = TypeVar('C_TB')
C_T_CDB = TypeVar('C_T_CDB', bound=CDataBase)
Self = TypeVar('Self')
P = ParamSpec('P')
R = TypeVar('R')


if TYPE_CHECKING:
    class PointerBase(Generic[C_T], _Pointer[C_T]):
        ...
else:
    class PointerBase(Generic[C_T], CDataBase):
        ...


class Pointer(Generic[C_T], PointerBase[C_T]):
    _type_: type[C_T]
    contents: C_T

    def __class_getitem__(cls, _type: C_T) -> type[PointerBound]:
        class PointerInnerClass(PointerBound):
            __bound_value__ = _type

        return PointerInnerClass

    @staticmethod
    def _norm_ptr(cls_type: C_T) -> Pointer[C_T]:
        ptr_type = POINTER(cls_type)  # type: ignore

        if hasattr(cls_type, '_ctypes_patch_getfunc'):
            from .ctypes import make_callback_returnable

            return make_callback_returnable(ptr_type)  # type: ignore

        return ptr_type  # type: ignore

    def __new__(cls: type[Self], cls_type: C_T | None = None) -> type[Pointer[C_T]]:
        from .struct import Struct

        if cls_type.__class__.__base__ is Struct:
            return cls_type  # type: ignore

        if cls_type is None:
            if not hasattr(cls, '_type_'):
                return c_void_p(None)  # type: ignore

            cls_type = cls._type_  # type: ignore

        return Pointer._norm_ptr(cls_type)

    @staticmethod
    def _to_normalize(value: Any) -> bool:
        return hasattr(value, 'mro') and (Pointer in value.mro())

    @staticmethod
    def normalize(value: Any) -> type[CDataBase]:
        from .utils import normalize_ctype

        value = normalize_ctype(value)

        if Pointer._to_normalize(value):
            if hasattr(value, '__args__'):
                basecls = value.__args__[0]
            else:
                basecls = value.__bound_value__

            basecls = normalize_ctype(basecls)
            if Pointer._to_normalize(basecls):
                basecls = Pointer.normalize(basecls)

            value = Pointer._norm_ptr(basecls)

        return normalize_ctype(value)


class PointerBound(Pointer):  # type: ignore
    __bound_value__: C_TB  # type: ignore

    def __new__(cls: type[Self], size: int | None = None) -> Pointer[C_TB]:
        from .utils import normalize_ctype

        nvalue = normalize_ctype(cls.__bound_value__)  # type: ignore
        bvalue = nvalue if size is None else nvalue * size

        return Pointer._norm_ptr(bvalue)()  # type: ignore


class FuncPointer(Generic[P, R]):
    restype: type[CDataBase]
    argtypes: Sequence[type[CDataBase]]
    errcheck: Callable[[type[CDataBase] | None, FuncPointer[P, R], tuple[CDataBase, ...]], CDataBase]

    def __new__(cls: type[Self], func: c_void_p | Callable[P, R]) -> Self:
        ...

    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> R:
        ...


class CallingConvention(str, Enum):
    C = 'cdecl'
    Win32 = 'stdcall'

    def __call__(self, func: F) -> F:
        if not callable(func):
            raise ValueError(
                'CallingConvention: Call must be used as a decorator on a funtion!'
            )

        func.__dict__.__setitem__('__ctdffi_cconv__', self)

        return func


class MetaClassDictBase(dict[str, Any]):
    @abstractmethod
    def _setitem_(self, name: str, value: object, /) -> None:
        ...

    def __setitem__(self, name: str, value: object, /) -> None:
        if name.startswith('__'):
            return dict.__setitem__(self, name, value)

        return self._setitem_(name, value)



from . import ctypes  # noqa: F401, F403
from .cython import *  # noqa: F401, F403
from .library import *  # noqa: F401, F403
from .libs import *  # noqa: F401, F403
from .string import *  # noqa: F401, F403
from .struct import *  # noqa: F401, F403
from .types import *  # noqa: F401, F403
from .utils import *  # noqa: F401, F403



from __future__ import annotations

import array
import mmap
import pickle
import sys
from abc import abstractmethod
from ctypes import POINTER, Array, Union, c_char, c_char_p, cast
from typing import (
    TYPE_CHECKING, Any, Iterable, NoReturn, Protocol, SupportsIndex, TypeAlias, TypeVar, runtime_checkable
)

from .types import CDataBase

# From ctypesgen.printer_python.preamble.3_2, which isn't importable. Added typing and did small fixes.


__all__ = [
    'StrType',
    'UserString',
    'MutableString',
    'String'
]


@runtime_checkable
class SupportsString(Protocol):
    @abstractmethod
    def __str__(self) -> str:
        pass


ReadOnlyBuffer: TypeAlias = bytes
if TYPE_CHECKING:
    WriteableBuffer: TypeAlias = bytearray | memoryview | array.array[Any] | mmap.mmap | CDataBase | pickle.PickleBuffer
else:
    WriteableBuffer: TypeAlias = bytearray | memoryview | array.array | mmap.mmap | CDataBase | pickle.PickleBuffer

ReadableBuffer: TypeAlias = ReadOnlyBuffer | WriteableBuffer


def _args_nn(val: Any) -> tuple[Any, ...]:
    return () if val is None else (val, )


class UserString:
    data: bytes

    def __init__(self, seq: bytes | UserString | SupportsString) -> None:
        if isinstance(seq, bytes):
            self.data = seq
        elif isinstance(seq, UserString):
            self.data = seq.data[:]
        else:
            self.data = str(seq).encode()

    def __bytes__(self) -> bytes:
        return self.data

    def __str__(self) -> str:
        return self.data.decode()

    def __repr__(self) -> str:
        return repr(self.data)

    def __int__(self) -> int:
        return int(self.data.decode())

    def __long__(self) -> int:
        return int(self.data.decode())

    def __float__(self) -> float:
        return float(self.data.decode())

    def __complex__(self) -> complex:
        return complex(self.data.decode())

    def __hash__(self) -> int:
        return hash(self.data)

    def __le__(self, string: bytes | UserString | SupportsString) -> bool:
        if isinstance(string, bytes):
            return self.data <= string
        elif isinstance(string, UserString):
            return self.data <= string.data
        else:
            return self.data <= str(string).encode()

    def __lt__(self, string: bytes | UserString | SupportsString) -> bool:
        if isinstance(string, bytes):
            return self.data < string
        elif isinstance(string, UserString):
            return self.data < string.data
        else:
            return self.data < str(string).encode()

    def __ge__(self, string: bytes | UserString | SupportsString) -> bool:
        if isinstance(string, bytes):
            return self.data >= string
        elif isinstance(string, UserString):
            return self.data >= string.data
        else:
            return self.data >= str(string).encode()

    def __gt__(self, string: bytes | UserString | SupportsString) -> bool:
        if isinstance(string, bytes):
            return self.data > string
        elif isinstance(string, UserString):
            return self.data > string.data
        else:
            return self.data > str(string).encode()

    def __eq__(self, string: bytes | UserString | SupportsString) -> bool:
        if isinstance(string, bytes):
            return self.data == string
        elif isinstance(string, UserString):
            return self.data == string.data
        else:
            return self.data == str(string).encode()

    def __ne__(self, string: bytes | UserString | SupportsString) -> bool:
        if isinstance(string, bytes):
            return self.data != string
        elif isinstance(string, UserString):
            return self.data != string.data
        else:
            return self.data != str(string).encode()

    def __contains__(self, char: SupportsIndex | bytes) -> bool:
        return char in self.data

    def __len__(self) -> int:
        return len(self.data)

    def __getitem__(self: C_USB, index: int) -> C_USB:
        return self.__class__(self.data[index])

    def __getslice__(self: C_USB, start: int, end: int) -> C_USB:
        start = max(start, 0)
        end = max(end, 0)
        return self.__class__(self.data[start:end])

    def __add__(self: C_USB, other: bytes | UserString | SupportsString) -> C_USB:
        if isinstance(other, UserString):
            return self.__class__(self.data + other.data)
        elif isinstance(other, bytes):
            return self.__class__(self.data + other)
        else:
            return self.__class__(self.data + str(other).encode())

    def __radd__(self: C_USB, other: bytes | UserString | SupportsString) -> C_USB:
        if isinstance(other, UserString):
            return self.__class__(other.data + self.data)
        elif isinstance(other, bytes):
            return self.__class__(other + self.data)
        else:
            return self.__class__(str(other).encode() + self.data)

    def __mul__(self: C_USB, n: SupportsIndex) -> C_USB:
        return self.__class__(self.data * n)

    __rmul__ = __mul__

    def __mod__(self: C_USB, args: Any) -> C_USB:
        return self.__class__(self.data % args)

    # the following methods are defined in alphabetical order:
    def capitalize(self: C_USB) -> C_USB:
        return self.__class__(self.data.capitalize())

    def center(self: C_USB, width: SupportsIndex, fillchar: bytes | None = None) -> C_USB:
        return self.__class__(self.data.center(width, *_args_nn(fillchar)))

    def count(
        self: C_USB, sub: ReadableBuffer | SupportsIndex,
        start: SupportsIndex | None = 0, end: SupportsIndex | None = sys.maxsize
    ) -> int:
        return self.data.count(sub, start, end)

    def decode(self: C_USB, encoding: str | None = None, errors: str | None = None) -> C_USB:
        args = (encoding, errors) if (encoding and errors) else (encoding, ) if encoding else ()
        return self.__class__(self.data.decode(*args))

    def encode(self: C_USB, encoding: str | None = None, errors: str | None = None) -> C_USB:
        args = (encoding, errors) if (encoding and errors) else (encoding, ) if encoding else ()
        return self.__class__(self.data.decode(*args))

    def endswith(
        self: C_USB, suffix: ReadableBuffer | tuple[ReadableBuffer, ...],
        start: SupportsIndex | None = 0, end: SupportsIndex | None = sys.maxsize
    ) -> bool:
        return self.data.endswith(suffix, start, end)

    def expandtabs(self: C_USB, tabsize: SupportsIndex = 8) -> C_USB:
        return self.__class__(self.data.expandtabs(tabsize))

    def find(
        self: C_USB, sub: ReadableBuffer | SupportsIndex,
        start: SupportsIndex | None = 0, end: SupportsIndex | None = sys.maxsize
    ) -> int:
        return self.data.find(sub, start, end)

    def index(
        self: C_USB, sub: ReadableBuffer | SupportsIndex,
        start: SupportsIndex | None = 0, end: SupportsIndex | None = sys.maxsize
    ) -> int:
        return self.data.index(sub, start, end)

    def isalpha(self: C_USB) -> bool:
        return self.data.isalpha()

    def isalnum(self: C_USB) -> bool:
        return self.data.isalnum()

    def isdecimal(self: C_USB) -> bool:
        return str(self.data).isdecimal()

    def isdigit(self: C_USB) -> bool:
        return self.data.isdigit()

    def islower(self: C_USB) -> bool:
        return self.data.islower()

    def isnumeric(self: C_USB) -> bool:
        return str(self.data).isnumeric()

    def isspace(self: C_USB) -> bool:
        return self.data.isspace()

    def istitle(self: C_USB) -> bool:
        return self.data.istitle()

    def isupper(self: C_USB) -> bool:
        return self.data.isupper()

    def join(self: C_USB, seq: Iterable[ReadableBuffer]) -> C_USB:
        return self.__class__(self.data.join(seq))

    def ljust(self: C_USB, width: SupportsIndex, fillchar: bytes | bytearray | None = None) -> C_USB:
        return self.__class__(self.data.ljust(width, *_args_nn(fillchar)))

    def lower(self: C_USB) -> C_USB:
        return self.__class__(self.data.lower())

    def lstrip(self: C_USB, chars: ReadableBuffer | None = None) -> C_USB:
        return self.__class__(self.data.lstrip(chars))

    def partition(self: C_USB, sep: ReadableBuffer) -> tuple[bytes, bytes, bytes]:
        return self.data.partition(sep)

    def replace(self: C_USB, old: ReadableBuffer, new: ReadableBuffer, maxsplit: SupportsIndex = -1) -> C_USB:
        return self.__class__(self.data.replace(old, new, maxsplit))

    def rfind(
        self: C_USB, sub: ReadableBuffer | SupportsIndex,
        start: SupportsIndex | None = 0, end: SupportsIndex | None = sys.maxsize
    ) -> int:
        return self.data.rfind(sub, start, end)

    def rindex(
        self: C_USB, sub: ReadableBuffer | SupportsIndex,
        start: SupportsIndex | None = 0, end: SupportsIndex | None = sys.maxsize
    ) -> int:
        return self.data.rindex(sub, start, end)

    def rjust(self: C_USB, width: SupportsIndex, fillchar: bytes | bytearray | None = None) -> C_USB:
        return self.__class__(self.data.rjust(width, *_args_nn(fillchar)))

    def rpartition(self: C_USB, sep: ReadableBuffer) -> tuple[bytes, bytes, bytes]:
        return self.data.rpartition(sep)

    def rstrip(self: C_USB, chars: ReadableBuffer | None = None) -> C_USB:
        return self.__class__(self.data.rstrip(chars))

    def split(self, sep: ReadableBuffer | None, maxsplit: SupportsIndex = -1) -> list[bytes]:
        return self.data.split(sep, maxsplit)

    def rsplit(self, sep: ReadableBuffer | None, maxsplit: SupportsIndex = -1) -> list[bytes]:
        return self.data.rsplit(sep, maxsplit)

    def splitlines(self: C_USB, keepends: bool = False) -> list[bytes]:
        return self.data.splitlines(keepends)

    def startswith(
        self, prefix: ReadableBuffer | tuple[ReadableBuffer, ...],
        start: SupportsIndex | None = 0, end: SupportsIndex | None = sys.maxsize
    ) -> bool:
        return self.data.startswith(prefix, start, end)

    def strip(self: C_USB, chars: ReadableBuffer | None = None) -> C_USB:
        return self.__class__(self.data.strip(chars))

    def swapcase(self: C_USB) -> C_USB:
        return self.__class__(self.data.swapcase())

    def title(self: C_USB) -> C_USB:
        return self.__class__(self.data.title())

    def translate(self: C_USB, table: ReadableBuffer | None, /, delete: bytes | None = None) -> C_USB:
        return self.__class__(self.data.translate(table, *_args_nn(delete)))

    def upper(self: C_USB) -> C_USB:
        return self.__class__(self.data.upper())

    def zfill(self: C_USB, width: SupportsIndex) -> C_USB:
        return self.__class__(self.data.zfill(width))


C_USB = TypeVar('C_USB', bound=UserString)


class MutableString(UserString):
    def __init__(self: C_MPB, string_data: str = '') -> None:
        self.data = str(string_data).encode()

    def __hash__(self) -> NoReturn:
        raise TypeError("unhashable mutable type")

    def __setitem__(self, index: int, sub: UserString | bytes | SupportsString) -> None:
        if index < 0:
            index += len(self.data)

        if index < 0 or index >= len(self.data):
            raise IndexError

        self.data = self.data[:index] + sub + self.data[index + 1:]  # type: ignore

    def __delitem__(self, index: int) -> None:
        if index < 0:
            index += len(self.data)

        if index < 0 or index >= len(self.data):
            raise IndexError

        self.data = self.data[:index] + self.data[index + 1:]

    def __setslice__(self, start: int, end: int, sub: UserString | bytes | SupportsString) -> None:
        start = max(start, 0)
        end = max(end, 0)

        if isinstance(sub, UserString):
            self.data = self.data[:start] + sub.data + self.data[end:]
        elif isinstance(sub, bytes):
            self.data = self.data[:start] + sub + self.data[end:]
        else:
            self.data = self.data[:start] + str(sub).encode() + self.data[end:]

    def __delslice__(self, start: int, end: int) -> None:
        start = max(start, 0)
        end = max(end, 0)
        self.data = self.data[:start] + self.data[end:]

    def __iadd__(self: C_MPB, other: UserString | bytes | SupportsString) -> C_MPB:
        if isinstance(other, UserString):
            self.data += other.data
        elif isinstance(other, bytes):
            self.data += other
        else:
            self.data += str(other).encode()
        return self

    def __imul__(self: C_MPB, n: SupportsIndex) -> C_MPB:
        self.data *= n
        return self

    def immutable(self) -> UserString:
        return UserString(self.data)


C_MPB = TypeVar('C_MPB', bound=MutableString)


class String(MutableString, Union):
    _fields_ = [("raw", POINTER(c_char)), ("data", c_char_p)]

    def __init__(self, obj: Any = "") -> None:
        if isinstance(obj, (bytes, UserString)):
            self.data = bytes(obj)
        else:
            self.raw = obj

    def __len__(self) -> int:
        return self.data and len(self.data) or 0

    @classmethod
    def from_param(cls, obj: StrType | int | None) -> StrType:
        if obj is None or obj == 0:  # Convert None or 0
            return cls(POINTER(c_char)())  # type: ignore
        elif isinstance(obj, bytes):  # Convert from bytes
            return cls(obj)
        elif isinstance(obj, str):  # Convert from str
            return cls(obj.encode())
        elif isinstance(obj, int):  # Convert from raw pointer
            return cls(cast(obj, POINTER(c_char)))
        elif isinstance(obj, String):  # Convert from String
            return obj
        elif isinstance(obj, c_char_p):  # Convert from c_char_p
            return obj
        elif isinstance(obj, POINTER(c_char)):  # Convert from POINTER(c_char)
            return obj
        elif isinstance(obj, c_char * len(obj)):  # Convert from c_char array
            return obj

        # Convert from object
        return String.from_param(obj._as_parameter_)


StrType = str | String | c_char_p | Array[c_char] | POINTER(c_char)



from __future__ import annotations

from ctypes import Structure
from inspect import get_annotations
from typing import TYPE_CHECKING, Any, Generic, Mapping, TypeVar

from .ctypes import make_callback_returnable
from .types import CDataBase, MetaClassDictBase, Self, StructMetaBase
from .utils import _protected_keys, as_cfunc

__all__ = [
    'StructMeta', 'Struct', 'OpaqueStruct'
]


class StructMetaDict(MetaClassDictBase):
    def _setitem_(self, name: str, value: Any, /) -> None:
        if callable(value):
            value = as_cfunc(value, name)

            self['__slots__'].append(name)
            self['_fields_'].append((name, value))

            return

        return dict.__setitem__(self, name, value)


class StructMeta(StructMetaBase):
    __bases__: tuple[type, ...]
    __slots__: list[str]
    _fields_: list[tuple[str, CDataBase]]  # type: ignore

    @classmethod
    def __prepare__(metacls, name: str, bases: tuple[type, ...], /, **kwargs: Any) -> Mapping[str, object]:
        return StructMetaDict(__slots__=[], _fields_=[])


class StructureBase(Generic[Self]):
    if TYPE_CHECKING:
        def __init__(self: type[Self]) -> None:  # type: ignore
            ...

    @staticmethod
    def annotate(cls: C_STB) -> C_STB:
        if Struct not in cls.mro():
            raise ValueError(
                'Struct.annotate: The annotated class must inherit from Struct!'
            )

        for key, value in get_annotations(cls, eval_str=True).items():
            if key.startswith('__') or key in _protected_keys:
                continue

            cls.__slots__.append(key)
            cls._fields_.append((key, value))

        return make_callback_returnable(cls)  # type: ignore


class Struct(StructureBase, Structure, metaclass=StructMeta):  # type: ignore
    ...


C_STB = TypeVar('C_STB', bound=StructMeta)


class OpaqueStruct(Struct):
    def __init__(self) -> None:
        raise NotImplementedError



from __future__ import annotations

from types import ModuleType
from typing import Any, Mapping, NoReturn

from .libs import PyCapsule
from .struct import Struct, StructMeta
from .types import MetaClassDictBase, Self
from .utils import as_cfunc, normalize_cfunc

__all__ = [
    'CythonModuleMeta', 'CythonModule'
]

_DUMMYMODULE = object()


class CythonModuleMetaDict(MetaClassDictBase):
    def __init__(self, cls_name: str, module: ModuleType | None) -> None:
        self.cls_name = cls_name
        self.module = module

        if module:
            self.capsules = module.__pyx_capi__
        else:
            self.capsules = dict[str, Any]()

        super().__init__(module=self.module, capsules=self.capsules)

    def _setitem_(self, name: str, value: Any, /) -> None:
        if callable(value):
            if self.module is None:
                value = self._raise_module_unavailable
            else:
                norm = normalize_cfunc(value, name)
                capsule_name = norm.oname or norm.name

                try:
                    capsule = self.capsules[capsule_name]
                except KeyError as e:
                    raise AttributeError(capsule_name) from e

                func_type = as_cfunc(norm)

                mangled_name = PyCapsule.GetName(capsule)
                capsule_ptr = PyCapsule.GetPointer(capsule, mangled_name)

                value = func_type(capsule_ptr)

        return dict.__setitem__(self, name, value)

    def _raise_module_unavailable(self, *args: Any, **kwargs: Any) -> NoReturn:
        raise ModuleNotFoundError(
            f'{self.cls_name}: The cython module is unavailable!'
        )


class CythonModuleMeta(StructMeta):
    @classmethod
    def __prepare__(metacls, name: str, bases: tuple[type, ...], /, **kwargs: Any) -> Mapping[str, object]:
        if 'module' not in kwargs:
            raise ValueError(
                'CythonModule: You must specify a module with subclass kwargs!'
            )

        module = kwargs.pop('module')

        if module is _DUMMYMODULE:
            return dict[str, Any]()

        if module is not None and not hasattr(module, '__pyx_capi__'):
            raise ValueError(
                'CythonModule: Passed module isn\'t a cython module!'
            )

        return CythonModuleMetaDict(name, module)

    def __new__(
        cls: type[Self], name: str, bases: tuple[type, ...], namespace: dict[str, Any], /, **kwargs: Any
    ) -> Self:
        return super().__new__(cls, name, bases, namespace)  # type: ignore


class CythonModule(Struct, metaclass=CythonModuleMeta, module=_DUMMYMODULE):
    ...



from __future__ import annotations

from typing import Any, Mapping, cast

from ctypesgen.libraryloader import LibraryLoader, load_library  # type: ignore

from .types import CallingConvention, MetaClassDictBase, Self
from .utils import normalize_cfunc

__all__ = [
    'LibraryMeta', 'Library',
    'CallingConvention'
]


class LibraryMetaDict(MetaClassDictBase):
    def __init__(self, lib_name: str, def_cconv: CallingConvention):
        self.lib = load_library(lib_name)
        self.def_cconv = def_cconv
        self['__pytydffi_lib__'] = self.lib

    def _setitem_(self, name: str, value: Any, /) -> None:
        if callable(value):
            norm = normalize_cfunc(value, name, self.def_cconv)

            value = self.lib.get(norm.oname or norm.name, norm.cconv.value)
            value.argtypes = norm.oargs_types or norm.args_types
            value.restype = norm.ores_type or norm.res_type

        return dict.__setitem__(self, name, value)


class LibraryMeta(type):
    @classmethod
    def _check_self(self, name: str, bases: tuple[type, ...], **kwargs: Any) -> str | None:
        if name == 'Library' and not bases:
            return None

        if Library not in bases:
            raise ValueError(
                'LibraryMeta: You should inherit from Library, not LibraryMeta directly!'
            )

        lib_name = kwargs.get('lib', None)

        if lib_name is None:
            raise ValueError(
                'Library: You have to specify the library name with `lib=\'libname\'`!'
            )

        return cast(str, lib_name)

    @property
    def lib(self) -> LibraryLoader:
        return self.__dict__.__getitem__('__pytydffi_lib__')

    @classmethod
    def __prepare__(metacls, name: str, bases: tuple[type, ...], /, **kwargs: Any) -> Mapping[str, object]:
        lib_name = LibraryMeta._check_self(name, bases, **kwargs)

        if lib_name is None:
            return dict()

        return LibraryMetaDict(lib_name, kwargs.get('cconv', CallingConvention.C))

    def __new__(
        cls: type[Self], name: str, bases: tuple[type, ...], namespace: dict[str, Any],
        /, **kwargs: Any
    ) -> Self:
        LibraryMeta._check_self(name, bases, **kwargs)

        return type.__new__(cls, name, bases, namespace)


class Library(metaclass=LibraryMeta):
    ...




import ctypes
from ctypes import (
    CFUNCTYPE, POINTER, PYFUNCTYPE, Array, Structure, _SimpleCData, addressof, c_char, c_char_p, c_double, c_float,
    c_int, c_int8, c_int16, c_int32, c_int64, c_longlong, c_short, c_size_t, c_ssize_t, c_ubyte, c_uint, c_uint8,
    c_uint16, c_uint32, c_uint64, c_ulonglong, c_ushort, c_void_p, memmove
)
from ctypes import py_object as py_object_t
from ctypes import pythonapi, sizeof, string_at
from functools import lru_cache
from types import FunctionType, MappingProxyType
from typing import TYPE_CHECKING, Any, TypeVar, cast

from .string import String
from .types import C_T_CDB, CDataBase

if TYPE_CHECKING:
    from ctypes import _Pointer  # type: ignore
else:
    _Pointer = None


__all__ = [
    'c_void_p',

    'c_char', 'c_char_p', 'StrType',

    'c_float', 'c_double',

    'c_int',
    'c_int8', 'c_int16', 'c_int32', 'c_int64',
    'c_int8_t', 'c_int16_t', 'c_int32_t', 'c_int64_t',

    'c_uint',
    'c_uint8', 'c_uint16', 'c_uint32', 'c_uint64',
    'c_uint8_t', 'c_uint16_t', 'c_uint32_t', 'c_uint64_t',

    'c_size_t', 'c_ptrdiff_t', 'c_intptr_t',

    'py_object_t', 'py_object', 'CFUNCTYPE', 'FuncPointerType',

    'PyTypeObject', 'PyObject', 'PyVarObject', 'PyDictObject', 'StgDictObject',

    'mappingproxyobject', 'ffi_type',

    'None_ptr',

    'get_stgdict_of_type', 'make_callback_returnable',
]


c_int8_t = c_char
c_int16_t = c_short
c_int32_t = c_int
c_int64_t = c_longlong
c_uint8_t = c_ubyte
c_uint16_t = c_ushort
c_uint32_t = c_uint
c_uint64_t = c_ulonglong

if TYPE_CHECKING:
    c_ptrdiff_t: type[_SimpleCData[int]]
    c_intptr_t: type[_SimpleCData[int]]
    from ctypes import _FuncPointer as FuncPointerType
    py_object = py_object_t[Any]
else:
    FuncPointerType = FunctionType
    py_object = py_object_t


_int_types = (c_int16, c_int32, c_int64)

for t in _int_types:
    if sizeof(t) == sizeof(c_size_t):
        c_ptrdiff_t = c_intptr_t = t

StrType = str | Array[c_char] | c_char_p | String


# The PyTypeObject struct from 'Include/object.h'.
# This is a forward declaration, fields are set later once PyVarObject has been declared.


class PyTypeObject(Structure):
    ...


# The PyObject struct from 'Include/object.h'.
class PyObject(Structure):
    _fields_ = [
        ('ob_refcnt', c_ssize_t),
        ('ob_type', POINTER(PyTypeObject)),
    ]


# The PyVarObject struct from 'Include/object.h'.
class PyVarObject(Structure):
    _fields_ = [
        ('ob_base', PyObject),
        ('ob_size', c_ssize_t),
    ]


# This structure is not stable across Python versions, but the few fields that we use probably won't change.
PyTypeObject._fields_ = [
    ('ob_base', PyVarObject),
    ('tp_name', c_char_p),
    ('tp_basicsize', c_ssize_t),
    ('tp_itemsize', c_ssize_t),
    # There are many more fields, but we're only interested in the size fields, so we can leave out everything else.
]


# The PyTypeObject structure for the dict class.
# This is used to determine the size of the PyDictObject structure.
PyDict_Type = PyTypeObject.from_address(id(dict))


# The PyDictObject structure from 'Include/dictobject.h'.
# This structure is not stable across Python versions, and did indeed change in recent Python releases.
# Because we only care about the size of the structure and not its actual contents,
# we can declare it as an opaque byte array, with the length taken from PyDict_Type.
class PyDictObject(Structure):
    _fields_ = [
        ('PyDictObject_opaque', (c_ubyte * PyDict_Type.tp_basicsize)),
    ]


# The mappingproxyobject struct from 'Objects/descrobject.c'.
# This structure is not officially stable across Python versions, but its layout hasn't changed since 2001.
class mappingproxyobject(Structure):
    _fields_ = [
        ('ob_base', PyObject),
        ('mapping', py_object),
    ]


# https://github.com/python/cpython/blob/main/Modules/_ctypes/libffi_osx/include/ffi.h#L140-L145
class ffi_type(Structure):
    ...


ffi_type._fields_ = [
    ('size', c_size_t),
    ('alignment', c_ushort),
    ('type', c_ushort),
    ('elements', POINTER(POINTER(ffi_type)))
]


None_ptr = ctypes.cast(id(None), POINTER(PyObject))

# https://github.com/python/cpython/blob/main/Modules/_ctypes/ctypes.h#L31-L32
GETFUNC = PYFUNCTYPE(py_object, c_void_p, c_ssize_t)
SETFUNC = PYFUNCTYPE(c_void_p, c_void_p, py_object, c_ssize_t)
# PARAMFUNC = PYFUNCTYPE(POINTER(PyCArgObject), CDataObject)
PARAMFUNC = PYFUNCTYPE(POINTER(PyObject), PyObject)


class CDataBaseFix(CDataBase):
    __module__: str
    __name__: str
    _actual_size: int
    _ctypes_patch_getfunc: GETFUNC
    _ctypes_patch_setfunc: SETFUNC


C_T_CBF = TypeVar('C_T_CBF', bound=CDataBaseFix)


# https://github.com/python/cpython/blob/main/Modules/_ctypes/ctypes.h#L200-L233
class StgDictObject(Structure):
    _fields_ = [
        ('dict', PyDictObject),
        ('size', c_ssize_t),
        ('align', c_ssize_t),
        ('length', c_ssize_t),

        ('ffi_type_pointer', ffi_type),

        ('proto', py_object),

        ('setfunc', SETFUNC),
        ('getfunc', GETFUNC),
        ('paramfunc', PARAMFUNC),

        ('argtypes', POINTER(PyObject)),
        ('converters', POINTER(PyObject)),
        ('restype', POINTER(PyObject)),

        ('checker', POINTER(PyObject)),
        ('flags', c_int),

        ('format', c_char_p),
        ('ndim', c_int),
        ('shape', POINTER(c_ssize_t)),
        ('strides', POINTER(c_ssize_t)),
        ('suboffsets', POINTER(c_ssize_t))
    ]


pythonapi.Py_IncRef.restype = None
pythonapi.Py_IncRef.argtypes = [POINTER(PyObject)]


def get_stgdict_of_type(tp: C_T_CDB) -> int:
    tptp = type(tp)

    if not isinstance(tp, type):
        raise TypeError(f'Expected a type object, not {tptp.__module__}.{tptp.__qualname__}')

    stgdict = tp.__dict__  # type: ignore[unreachable]
    if isinstance(stgdict, MappingProxyType):
        stgdict = mappingproxyobject.from_address(id(stgdict)).mapping

    dict_tp = type(stgdict)

    if dict_tp.__name__ != 'StgDict':
        raise TypeError(
            f'The given type\'s dict must be a StgDict, not {dict_tp.__module__}.{dict_tp.__qualname__}'
        )

    return id(stgdict)


@lru_cache
def _check_size(func_name: str, ctype: CDataBaseFix, size: c_size_t) -> int:
    if int(size) not in {0, ctype._actual_size}:
        raise ValueError(
            f'{func_name} for ctype {ctype}: Requested size {size} does not match actual size {ctype._actual_size}'
        )

    return ctype._actual_size


def make_callback_returnable(ctype: CDataBase) -> C_T_CBF:
    ctypef = cast(C_T_CBF, ctype)

    if hasattr(ctype, '_ctypes_patch_getfunc'):
        return ctypef

    stgdict_c = StgDictObject.from_address(get_stgdict_of_type(ctypef))

    for func_type in {'getfunc', 'setfunc'}:
        if ctypes.cast(getattr(stgdict_c, func_type), c_void_p).value is not None:
            raise ValueError(
                f'The ctype {ctypef.__module__}.{ctypef.__name__} already has a {func_type}'
            )

    @GETFUNC  # type: ignore
    def getfunc(ptr: c_void_p, size: c_size_t) -> py_object:
        return ctype.from_buffer_copy(string_at(ptr, _check_size('getfunc', ctype, size)))  # type: ignore

    @SETFUNC  # type: ignore
    def setfunc(ptr: c_void_p, value: py_object, size: c_size_t) -> c_void_p:
        memmove(ptr, addressof(value), _check_size('setfunc', ctype, size))
        pythonapi.Py_IncRef(None_ptr)

        return ctypes.cast(None_ptr, c_void_p).value  # type: ignore

    ctypef._actual_size = sizeof(ctype)
    ctypef._ctypes_patch_getfunc = getfunc
    ctypef._ctypes_patch_setfunc = setfunc

    stgdict_c.getfunc = getfunc
    stgdict_c.setfunc = setfunc

    return ctypef
